from sklearn.base import clone
from sklearn.metrics.scorer import check_scoring
from sklearn.utils import check_X_y
from sklearn.model_selection import check_cv
from time import time
from copy import deepcopy
import numpy as np

# TODO: make this use parallelism


def run_cv(clf, X, y, params, scoring='accuracy', cv=5, refit_best=True):
    """
    Runs cross-validation on a classifier.
    This function is designed for classifiers where we can quickly  compute
    the whole tuning path if we are allowed to pre-compute some quantities
    based on X before fitting the model.

    The clf object should be a sklearn compatible classifier
    with an additional clf.cv_init(X) function which does some precomputation
    based on the X data before precomputing the whole path.

    Parameters
    ----------
    clf:

    X, y:

    params: dict of lists

    scoring:

    cv:

    Output
    ------
    best_params, best_scores, best_clf, agg_results, all_cv_results
    """

    X, y = check_X_y(X, y,
                     accept_sparse='csr',
                     dtype='numeric')

    scorer = check_scoring(estimator=clf, scoring=scoring)

    # init cross-validation generator
    cv = check_cv(cv)
    folds = list(cv.split(X, y))
    n_folds = len(folds)

    all_param_settings = DoL2LoD(params)
    n_settings = len(all_param_settings)

    # get tuning path for each fold
    all_cv_results = []
    for f_idx, (test, train) in enumerate(folds):
        X_train = X[train, :]
        y_train = y[train]
        X_test = X[test, :]
        y_test = y[test]

        # clone(clf) is critical to send a copy! otherwise clf gets modified
        fold_results = get_path_scores(clf=clone(clf),
                                       X_train=X_train,
                                       y_train=y_train,
                                       X_test=X_test,
                                       y_test=y_test,
                                       scorer=scorer,
                                       params=params)

        all_cv_results.append(fold_results)

    # aggregate metrics of interst over folds
    agg_results = {'params': all_param_settings}
    metric_keys = ['test_score', 'train_score', 'runtime']
    for metric in metric_keys:
        agg_results['mean_' + metric] = []
        agg_results['std_' + metric] = []

    for s in range(n_settings):
        for metric in metric_keys:

            # values of the metric for a given parameter setting across
            # all folds
            vals = [all_cv_results[f][metric][s] for f in range(n_folds)]

            agg_results['mean_' + metric].append(np.mean(vals))
            agg_results['std_' + metric].append(np.std(vals))

    # get the best tuning parameter setting
    idx_best = np.argmax(agg_results['mean_test_score'])
    best_params = all_param_settings[idx_best]
    best_score = agg_results['mean_test_score'][idx_best]

    if refit_best:
        # refit classifier on full training data with best parameters
        best_clf = clone(clf)
        best_clf = best_clf.set_params(**best_params)
        best_clf = clf.fit(X, y)
    else:
        best_clf = None

    return best_params, best_score, best_clf, agg_results, all_cv_results


def get_path_scores(clf, X_train, y_train, X_test, y_test, scorer, params):

    # initalize classifier before cross validation
    start_time = time()
    clf.cv_init(X_train)
    init_runtime = time() - start_time

    cv_results = {'params': [],
                  'train_score': [],
                  'test_score': [],
                  'runtime': [],
                  'init_time': init_runtime}

    # fit and score classifier for each parameter setting
    # TODO: parallelize here
    for param_setting in DoL2LoD(params):
        clf.set_params(**param_setting)

        start_time = time()
        clf.fit(X_train, y_train)
        runtime = time() - start_time

        tr_score = scorer(clf, X_train, y_train)
        tst_score = scorer(clf, X_test, y_test)

        cv_results['params'].append(param_setting)
        cv_results['runtime'].append(runtime)
        cv_results['train_score'].append(tr_score)
        cv_results['test_score'].append(tst_score)

    return cv_results


def listify(x):
    """
    Returns a list
    """
    if not hasattr(x, '__len__'):
        return [x]
    else:
        return x


def DoL2LoD(DL):
    """
    Converts a dict of lists to a list of dices
    """
    dl = deepcopy(DL)
    for k in dl.keys():
        dl[k] = listify(dl[k])

    return [dict(zip(dl, t)) for t in zip(*dl.values())]



import numpy as np
from copy import deepcopy

from sklearn.base import BaseEstimator
from sklearn.utils import check_X_y

from dwd.utils import pm1
from dwd.gen_dwd import V, V_grad
from dwd.kernel_utils import KernelClfMixin
from dwd.cv import run_cv


class KernGDWD(BaseEstimator, KernelClfMixin):
    """
    Kernel Generalized Distance Weighted Discrimination

    Solves the kernel gDWD problem using the MM algorithm derived in Wang and Zou, 2017.

    Primary reference: Another look at distance-weighted discrimination by Boxiang Wang and Hui Zou, 2017

    Note the tuning parameter lambd is on a different scale the parameter C which is used in the SOCP formulation.

    Parameters
    ----------
    lambd: float
        Tuning parameter for DWD.

    q: float
        Tuning parameter for generalized DWD (the exponent on the margin terms). When q = 1, gDWD is equivalent to DWD.

    kernel: str, callable(X, Y, **kwargs)
        The kernel to use.

    kernel_kws: dict
        Any key word arguments for the kernel.

    implicit_P: bool
        Whether to use the implicit P^{-1} gamma formulation (in the publication) or the explicit computation (in the arxiv version).
    """

    def __init__(self, lambd=1.0, q=1.0, kernel='linear',
                 kernel_kws={}, implicit_P=True):
        self.lambd = lambd
        self.q = q

        self.kernel = kernel
        self.kernel_kws = kernel_kws

        self.implicit_P = implicit_P

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target vector relative to X

        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.

        Returns
        -------
        self : object
        """
        # TODO: what to do about multi-class

        self.classes_ = np.unique(y)
        self._Xfit = X  # Store K so we can compute predictions

        K = self._compute_kernel(X)

        K_eig = self._get_K_eig()

        # fit DWD
        alpha, offset, obj_vals, c = \
            solve_gen_kern_dwd(K=K,
                               y=y,
                               lambd=self.lambd,
                               q=self.q,
                               alpha_init=None,
                               offset_init=None,
                               sample_weight=None,
                               implicit_P=self.implicit_P,
                               obj_tol=1e-5, max_iter=100,
                               K_eig=K_eig)

        self.intercept_ = offset.reshape(-1)
        self.dual_coef_ = alpha.reshape(1, -1)

        return self

    def cv_init(self, X):
        """
        Initializes the object before computing a cross-valiation.
        """
        # I don't love that we have to set this here
        self._Xfit = X

        # Warning: we compute the kernel twice -- any way around this
        # without messing up the SKlearn API too badly?
        K = self._compute_kernel(X)
        self._set_K_eig(K)
        return self

    def _set_K_eig(self, X):
        """
        Precomputes eigen decomposition of K matrix which makes
        cross-validation much faster.
        """
        self._K_eig = get_K_eig(X)

    def _get_K_eig(self):
        if hasattr(self, '_K_eig'):
            return self._K_eig
        else:
            return None


class KernGDWDCV(BaseEstimator, KernelClfMixin):
    """
    Fits kernel gDWD with cross-validation. gDWD cross-validation
    can be significnatly faster if certain quantities are precomputed.

    Parameters
    ----------
    lambd_vals: list of floats
        The lambda values to cross-validate over.

    q_vals: list of floats
        The q-values to cross validate over.

    kernel: str, callable
        The kernel to use.

    kern_kws_vals: list of dicts
        The kernel parameters to validate over.

    cv:
        How to perform cross-valdiation. See documetnation in sklearn.model_selection.GridSearchCV.

    scoring:
        What metric to use to score cross-validation. See documetnation in sklearn.model_selection.GridSearchCV.

    """
    def __init__(self,
                 lambd_vals=np.logspace(-2, 2, 10),
                 q_vals=np.logspace(-2, 2, 5),
                 kernel='linear',
                 kernel_kws_vals={},
                 cv=5, scoring='accuracy'):

        self.lambd_vals = lambd_vals
        self.q_vals = q_vals
        self.kernel = kernel
        self.kernel_kws_vals = kernel_kws_vals

        self.cv = cv
        self.scoring = scoring

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target vector relative to X

        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.
        Returns
        -------
        self : object
        """
        # TODO: what to do about multi-class

        # formatting
        # TODO: figure out what we should actually check
        X, y = check_X_y(X, y, accept_sparse='csr',
                         dtype='numeric')

        self.classes_ = np.unique(y)

        # run cross validation
        params = {'q': self.q_vals, 'lambd': self.lambd_vals,
                  'kernel_kws': self.kernel_kws_vals}

        best_params, best_score, best_clf, agg_results, all_cv_results = \
            run_cv(clf=KernGDWD(kernel=self.kernel),
                   X=X, y=y,
                   params=params,
                   scoring=self.scoring,
                   cv=self.cv,
                   refit_best=True)

        self.best_estimator_ = best_clf

        self.best_params_ = best_params
        self.best_score_ = best_score
        self.agg_cv_results_ = agg_results
        self.all_cv_results_ = all_cv_results

        self._Xfit = self.best_estimator_._Xfit
        self.intercept_ = self.best_estimator_.intercept_
        self.dual_coef_ = self.best_estimator_.dual_coef_

        self.decision_function = self.best_estimator_.decision_function

        return self


def solve_gen_kern_dwd(K, y, lambd, q=1,
                       alpha_init=None, offset_init=None,
                       sample_weight=None,
                       implicit_P=True,
                       obj_tol=1e-5, max_iter=100,
                       K_eig=None):

    """
    Solves the kernel gDWD problem using the MM algorithm derived in Wang and Zou, 2017.

    Parameters
    ----------
    K: array-like, (n_samples, n_samples)
        The kernel.

    y: array-like, (n_samples, )
        The vector of binary class labels.

    lambd: float
        Tuning parameter for DWD.

    q: float
        Tuning parameter for generalized DWD (the exponent on the margin terms). When q = 1, gDWD is equivalent to DWD.

    alpha_init, offset_init:
        Initial values to start the optimization algorithm from.

    sample_weight: None, array-like (n_samples,)
        Optional weight for samples.

    implicit_P: bool
        Whether to use the implicit P^{-1} gamma formulation (in the publication) or the explicit computation (in the arxiv version).

    obj_tol: float
        Stopping condition for difference between successive objective
        functions.

    max_iter: int
        Maximum number of iterations to perform.

    K_eig: None or (U, D)
        Optional. The precomputed eigendecomposition of K.
    """

    # argument checking and formatting
    if lambd < 0:
        raise ValueError("Penalty term must be positive; got (lambd=%r)"
                         % lambd)

    if q < 0:
        raise ValueError("Weight term must be positive; got (q=%r)" % q)

    # TODO: add sample weights
    if sample_weight is not None:
        raise NotImplementedError

    K, y = check_X_y(K, y,
                     accept_sparse='csr',
                     dtype='numeric')

    assert K.shape[0] == K.shape[1]  # K must be Kernel matrix.

    # convert y to +/- 1
    y = pm1(y)  # convert y to  y +/- 1

    n_samples = K.shape[0]
    M = (q + 1) ** 2 / q

    # precompute data
    if implicit_P:

        # precompute data needed to do implicit P^{-1} gamma
        if K_eig is not None:
            U, Lam = K_eig
            Lam = Lam.ravel()
            assert U.shape == (n_samples, n_samples)
            assert len(Lam) == n_samples
        else:
            U, Lam = get_K_eig(K)

        # see section 4.1 of Wang and Zou, 2017 for details
        pi = Lam ** 2 + (2 * n_samples * lambd / M) * Lam

        ULP = np.multiply(U, Lam * (1.0 / pi))
        v = ULP.dot(U.T.dot(np.ones(n_samples)))

        Ucs = U.sum(axis=0)
        carson = Ucs.dot(np.multiply(ULP, Lam).dot(Ucs))
        g = 1.0 / (n_samples - carson)

    else:
        raise NotImplementedError('explicit_P has a bug ')
        # the objective function explodes
        # TODO: :figure out what is going wrong

        # explicltly calculate P^{-1}
        colsum = K.sum(axis=0).reshape(-1, 1)
        frank = K.dot(K) + (2 * n_samples * lambd / M) * K
        P = np.bmat([[np.array([[n_samples]]), colsum.T],
                     [colsum, frank]]).A
        P_inv = np.linalg.inv(P)

    # initialize variables
    if alpha_init is None:
        alpha = np.random.normal(size=n_samples)
        alpha /= np.linalg.norm(alpha)
    else:
        alpha = alpha_init

    if offset_init is None:
        offset = 0.0
    else:
        offset = offset_init

    # store objective values
    obj_vals = []
    prev_obj = kern_dwd_obj(K, y, q, lambd, alpha, offset)
    obj_vals.append(prev_obj)

    for i in range(max_iter):

        # get step
        if implicit_P:
            alpha_step, offset_step = \
                get_step_implicit_P(K=K, y=y, q=q, lambd=lambd,
                                    alpha=alpha, offset=offset,
                                    U=U, Lam=Lam, pi=pi, v=v, g=g)
        else:
            alpha_step, offset_step = \
                get_step_explicit_P(K=K, y=y, q=q, lambd=lambd,
                                    alpha=alpha, offset=offset,
                                    P_inv=P_inv)

        # update parameters
        alpha = alpha - alpha_step
        offset = offset - offset_step

        # stopping condition
        current_obj = kern_dwd_obj(K, y, q, lambd, alpha, offset)
        obj_vals.append(current_obj)

        if np.abs(current_obj - prev_obj) < obj_tol:
            break
        else:
            prev_obj = deepcopy(current_obj)

    # tuning paramter for SOCP formulation
    c = c_from_lambd(K, lambd, q, alpha)

    return alpha, offset, obj_vals, c


def get_K_eig(K):
    """
    Computes the eigendecomposition of K.

    Parameters
    ----------
    K: (n_samples, n_samples)
        Kernel matrix.

    Output
    ------
    U, D

    U: (n_features, n_features)
        Eigenvectors of K.

    D: (n_features, )
        Eigenvalues of K in decending order.
    """

    Lam, U = np.linalg.eigh(K)
    Lam = Lam[::-1]  # sort evals in decending order
    U = U[:, ::-1]

    return U, Lam


def get_step_implicit_P(K, y, q, lambd, alpha, offset, U, Lam, pi, v, g):
    """
    Computes the step size for one step of the MM algorithm.
    See Wang and Zou, 2017 for details.
    """
    n_samples = K.shape[0]

    z = y * V_grad(y * (K.dot(alpha) + offset), q=q) / n_samples

    alice = z + 2 * lambd * alpha

    bob = np.insert(-v, 0, values=1)
    bob = (sum(z) - z.T.dot(K.dot(alice))) * bob

    cathy = np.multiply(U, Lam * (1.0 / pi)).dot(U.T.dot(alice))
    cathy = np.insert(cathy, 0, 0)

    P_inv_gamma = g * bob + cathy

    step = (n_samples * q / ((q + 1) ** 2)) * P_inv_gamma

    offset_step = step[0]
    alpha_step = step[1:]

    return alpha_step, offset_step


def get_step_explicit_P(K, y, q, lambd, alpha, offset, P_inv):
    """
    Computes the step size for one step of the MM algorithm.
    See Wang and Zou, 2017 for details.
    """
    n_samples = K.shape[0]

    z = y * V_grad(y * (K.dot(alpha) + offset), q=q) / n_samples

    gamma = K.dot(z) + 2 * lambd * K.dot(alpha)
    gamma = np.insert(gamma, 0, z.sum())

    step = (n_samples * q / (q + 1) ** 2) * P_inv @ gamma

    offset_step = step[0]
    alpha_step = step[1:]

    return alpha_step, offset_step


def kern_dwd_obj(K, y, q, lambd, alpha, offset):
    """
    Objective function for kernel DWD.
    """
    return np.mean(V(y * (K.dot(alpha) + offset), q=q) +
                   lambd * alpha.T.dot(K.dot(alpha)))


def c_from_lambd(K, lambd, q, alpha):
    """
    Gets the tuning paramter, C, for the SOCP formulation of DWD
    from lambda.
    """
    beta_norm = np.sqrt(alpha.T.dot(K.dot(alpha)))
    return ((q + 1) ** (q + 1) / q ** q) * beta_norm ** (q + 1)



import numpy as np
from sklearn.metrics.pairwise import pairwise_kernels

from sklearn.utils.extmath import safe_sparse_dot
from sklearn.base import ClassifierMixin
from sklearn.utils.validation import FLOAT_DTYPES
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


class KernelClfMixin(ClassifierMixin):
    """
    Mixin for kernel classifiers.
    """
    # def __init__(self, kernel, kernel_kws={}):
    #     # TODO: kernel centering
    #     self.kernel = kernel
    #     self.kernel_kws = kernel_kws

    def fit(self, X, y, sample_weights=None):

        # the fit function needs to set the following
        self.classes_ = np.unique(y)
        self.__Xfit = X
        self.intercept_ = None
        self.dual_coef_ = None
        raise NotImplementedError

    def _compute_kernel(self, X):
        """

        Parameters
        ----------
        X: array-like, shape (n_samples_test, n_features)
            A matrix of new data

        Returns
        -------
        K: array-like, shape (n_samples_train, n_samples_test)
        """

        if self.kernel == 'precomputed':
            return X

        elif callable(self.kernel):
            return self.kernel(X=self._Xfit,
                               Y=X,
                               kernel=self.kerel)

        elif isinstance(self.kernel, str):
            return pairwise_kernels(X=self._Xfit,
                                    Y=X,
                                    metric=self.kernel,
                                    **self.kernel_kws)

    def decision_function(self, X):
        """Predict confidence scores for samples.
        The confidence score for a sample is the signed distance of that
        sample to the hyperplane.
        Parameters
        ----------
        X : array_like or sparse matrix, shape (n_samples, n_features)
            Samples.
        Returns
        -------
        array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)
            Confidence scores per (sample, class) combination. In the binary
            case, confidence score for self.classes_[1] where >0 means this
            class would be predicted.
        """

        K = self._compute_kernel(X)

        scores = safe_sparse_dot(K.T, self.dual_coef_.T,
                                 dense_output=True) + self.intercept_
        return scores.ravel() if scores.shape[1] == 1 else scores

    def predict(self, X):
        """Predict class labels for samples in X.
        Parameters
        ----------
        X : array_like or sparse matrix, shape (n_samples, n_features)
            Samples.
        Returns
        -------
        C : array, shape [n_samples]
            Predicted class label per sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(np.int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]


class KernelScaler(BaseEstimator, TransformerMixin):
    """Center a kernel matrix
    Let K(x, z) be a kernel defined by phi(x)^T phi(z), where phi is a
    function mapping x to a Hilbert space. KernelScaler scales (i.e.,
    normalized to have zero norm) the data without explicitly computing phi(x).
    It is equivalent to centering phi(x) with
    sklearn.preprocessing.StandardScaler(with_mean=False).
    Read more in the :ref:`User Guide <kernel_centering>`.
    """

    def __init__(self):
        # Needed for backported inspect.signature compatibility with PyPy
        pass

    def fit(self, K, y=None):
        """Fit KernelCenterer
        Parameters
        ----------
        K : numpy array of shape [n_samples, n_samples]
            Kernel matrix.
        Returns
        -------
        self : returns an instance of self.
        """
        K = check_array(K, dtype=FLOAT_DTYPES)
        self.K_diag_ = np.diag(K)
        return self

    def transform(self, K, copy=True):
        """Center kernel matrix.
        Parameters
        ----------
        K : numpy array of shape [n_samples1, n_samples2]
            Kernel matrix.
        copy : boolean, optional, default True
            Set to False to perform inplace computation.
        Returns
        -------
        K_new : numpy array of shape [n_samples1, n_samples2]
        """
        # check_is_fitted(self, 'K_diag_')

        K = check_array(K, copy=copy, dtype=FLOAT_DTYPES)

        n = len(self.K_diag_)
        s = 1.0 / np.sqrt(self.K_diag_ / n)

        return np.multiply(np.multiply(s, K), s)

    @property
    def _pairwise(self):
        return True



import numpy as np


def pm1(y):
    """
    Converts binary label vector y into +/- 1s

    Parameters
    ----------
    y: array-like, (n_samples, )
        The original labels. Must have binary labels.

    Output
    ------
    y: array-like, (n_samples, )
        y with +/- 1 values. Note the positive label is determined
        by np.unique(y)[1].
    """
    y = np.array(y).reshape(-1)
    labels = np.unique(y)

    # check binary
    if len(labels) != 2:
        raise ValueError('y must have binary labels;'
                         ' found {} labels'.format(len(labels)))

    y_pm1 = np.ones(len(y))
    y_pm1[y == labels[0]] = -1
    return y_pm1



import numpy as np
from copy import deepcopy

from sklearn.base import BaseEstimator
from sklearn.linear_model.base import LinearClassifierMixin
from sklearn.utils import check_X_y

from dwd.utils import pm1
from dwd.cv import run_cv


class GenDWD(BaseEstimator, LinearClassifierMixin):
    """
    Generalized Distance Weighted Discrimination

    Solves the gDWD problem using the MM algorithm derived in Wang and Zou, 2017.

    Primary reference: Another look at distance-weighted discrimination by Boxiang Wang and Hui Zou, 2017

    Note the tuning parameter lambd is on a different scale the parameter C which is used in the SOCP formulation.

    Parameters
    ----------
    lambd: float
        Tuning parameter for DWD.

    q: float
        Tuning parameter for generalized DWD (the exponent on the margin terms). When q = 1, gDWD is equivalent to DWD.

    implicit_P: bool
        Whether to use the implicit P^{-1} gamma formulation (in the publication) or the explicit computation (in the arxiv version).

    """
    def __init__(self, lambd=1.0, q=1, implicit_P=True):  # TODO: solver args
        self.lambd = lambd
        self.q = q
        self.implicit_P = implicit_P

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target vector relative to X

        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.

        Returns
        -------
        self : object
        """
        # TODO: what to do about multi-class

        # formatting
        # TODO: figure out what we should actually check
        X, y = check_X_y(X, y, accept_sparse='csr',
                         dtype='numeric')

        self.classes_ = np.unique(y)

        P0_eig = self._get_P0_eig()

        # fit DWD
        self.coef_, self.intercept_, self.obj_vals_, self.C_ = \
            solve_gen_dwd(X=X, y=y,
                          lambd=self.lambd, q=self.q,
                          sample_weight=sample_weight,
                          P0_eig=P0_eig,
                          beta_init=None, offset_init=None,
                          obj_tol=1e-5, max_iter=100)

        self.coef_ = self.coef_.reshape(1, -1)
        self.intercept_ = self.intercept_.reshape(-1)

        return self

    def cv_init(self, X):
        """
        Initializes the object before computing a cross-valiation.
        """
        self._set_P0_eig(X)
        return self

    def _set_P0_eig(self, X):
        """
        Precomputes eigen decomposition of P0 matrix which makes
        cross-validation much faster.
        """
        self._P0_eig = get_P0_eig(X)

    def _get_P0_eig(self):
        if hasattr(self, '_P0_eig'):
            return self._P0_eig
        else:
            return None


class GenDWDCV(BaseEstimator, LinearClassifierMixin):
    """
    Fits Genralized DWD with cross-validation. gDWD cross-validation
    can be significnatly faster if certain quantities are precomputed.

    Parameters
    ----------
    lambd_vals: list of floats
        The lambda values to cross-validate over.

    q_vals: list of floats
        The q-values to cross validate over.

    cv:
        How to perform cross-valdiation. See documetnation in sklearn.model_selection.GridSearchCV.

    scoring:
        What metric to use to score cross-validation. See documetnation in sklearn.model_selection.GridSearchCV.

    """
    def __init__(self,
                 lambd_vals=np.logspace(-2, 2, 10),
                 q_vals=np.logspace(-2, 2, 5),
                 cv=5, scoring='accuracy'):
        self.lambd_vals = lambd_vals
        self.q_vals = q_vals

        self.cv = cv
        self.scoring = scoring

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target vector relative to X

        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.
        Returns
        -------
        self : object
        """
        # TODO: what to do about multi-class

        # formatting
        # TODO: figure out what we should actually check
        X, y = check_X_y(X, y, accept_sparse='csr',
                         dtype='numeric')

        self.classes_ = np.unique(y)

        # run cross validation
        params = {'q': self.q_vals, 'lambd': self.lambd_vals}
        best_params, best_score, best_clf, agg_results, all_cv_results = \
            run_cv(clf=GenDWD(),
                   X=X, y=y,
                   params=params,
                   scoring=self.scoring,
                   cv=self.cv,
                   refit_best=True)

        self.best_estimator_ = best_clf

        self.best_params_ = best_params
        self.best_score_ = best_score
        self.agg_cv_results_ = agg_results
        self.all_cv_results_ = all_cv_results

        self.coef_ = self.best_estimator_.coef_
        self.intercept_ = self.best_estimator_.intercept_

        return self


def solve_gen_dwd(X, y, lambd, q=1,
                  sample_weight=None,
                  beta_init=None, offset_init=None,
                  implicit_P=True,
                  obj_tol=1e-5, max_iter=100,
                  P0_eig=None):

    """
    Solves the kernel gDWD problem using the MM algorithm derived in Wang and Zou, 2017.

    Parameters
    ----------
    X: array-like, (n_samples, n_features)
        Input X data.

    y: array-like, (n_samples, )
        The vector of binary class labels.

    lambd: float
        Tuning parameter for DWD.

    q: float
        Tuning parameter for generalized DWD (the exponent on the margin terms). When q = 1, gDWD is equivalent to DWD.

    beta_init, offset_init:
        Initial values to start the optimization algorithm from.

    sample_weight: None, array-like (n_samples,)
        Optional weight for samples.

    implicit_P: bool
        Whether to implicitly calculate P^{-1} gamma. If P0_eig is precomputed
        this is much faster.

    obj_tol: float
        Stopping condition for difference between successive objective
        functions.

    max_iter: int
        Maximum number of iterations to perform.

    P0_eig: None, tuple of (U, D)
        Precomputed eigenvectors and eigenvalues of P0. Optional.
    """

    # argument checking and formatting
    if lambd < 0:
        raise ValueError("Penalty term must be positive; got (lambd=%r)"
                         % lambd)

    if q < 0:
        raise ValueError("Weight term must be positive; got (q=%r)" % q)

    # TODO: add sample weights
    if sample_weight is not None:
        raise NotImplementedError

    X, y = check_X_y(X, y,
                     accept_sparse='csr',
                     dtype='numeric')

    # convert y to +/- 1
    y = pm1(y)

    n_samples, n_features = X.shape
    M = (q + 1) ** 2 / q

    # initialize variables
    if beta_init is None:
        beta = np.random.normal(size=n_features)
    else:
        beta = beta_init

    if offset_init is None:
        offset = 0.0
    else:
        offset = offset_init

    # precompute data
    if implicit_P:
        # data needed to implicitly calculate P_inv @ gamma
        if P0_eig is not None:
            U, D = P0_eig
            D = D.ravel()
            assert U.shape == (n_features + 1, n_features + 1)
            assert len(D) == n_features + 1
        else:
            U, D = get_P0_eig(X)

        pi = D + 2 * n_samples * lambd / M

        u1 = U[0, :]
        v = np.multiply(U, 1.0 / pi).dot(u1)
        g = 2 * n_samples * q * lambd / \
            ((q + 1) ** 2 + 2 * n_samples * lambd * sum(v))

    else:
        # explicitly create P_inv matrix
        col_sums = X.sum(axis=0).reshape(-1, 1)
        d = X.T.dot(X) + (2 * n_samples * lambd / M) * np.eye(n_features)
        P = np.bmat([[np.array([[n_samples]]), col_sums.T],
                     [col_sums, d]]).A
        P_inv = np.linalg.inv(P)

    # store objective values
    obj_vals = []
    prev_obj = dwd_obj(X, y, q, lambd, beta, offset)
    obj_vals.append(prev_obj)

    for i in range(max_iter):

        # get step
        if implicit_P:
            beta_step, offset_step = \
                get_step_implicit(X, y, beta, offset, lambd, q,
                                  U, v, g, pi)
        else:
            beta_step, offset_step = \
                get_step_explicit(X, y, beta, offset, lambd, q, P_inv)

        offset = offset - offset_step  # step[0]
        beta = beta - beta_step  # step[1:]

        # stopping condition
        current_obj = dwd_obj(X, y, q, lambd, beta, offset)
        obj_vals.append(current_obj)

        if np.abs(current_obj - prev_obj) < obj_tol:
            break
        else:
            prev_obj = deepcopy(current_obj)

    c = c_from_lambd(lambd, q, beta)

    return beta, offset, obj_vals, c


def get_P0_eig(X):
    """
    Equation (3.6)

    Parameters
    ----------
    X: (n_samples, n_features)

    Output
    ------
    U, D

    U: (n_features, n_features)
        Eigenvectors of P0.

    D: (n_features, )
        Eigenvalues of P0 in decending order.
    """
    n = np.array([[X.shape[0]]])
    colsum = X.sum(axis=0).reshape(-1, 1)

    # create P0
    P0 = [[n, colsum.T],
          [colsum, X.T.dot(X)]]  # (d+1 x d+1)
    P0 = np.bmat(P0).A

    # compute eigen decomp of P0
    D, U = np.linalg.eigh(P0)
    D = D[::-1]  # sort evals in decending order
    U = U[:, ::-1]

    return U, D


def get_step_implicit(X, y, beta, offset, lambd, q, U, v, g, pi):
    """
    Gets the MM step by implicitly calculating P^{-1} gamma
    """
    n_samples = X.shape[0]

    # compute update
    z = y * V_grad(y * (X.dot(beta) + offset), q=q) / n_samples

    gamma = X.T.dot(z) + 2 * lambd * beta
    gamma = np.insert(gamma, 0, z.sum())

    # implictly compute P_inv @ gamma

    P_inv_gamma = np.multiply(U, 1.0 / pi).dot(U.T.dot(gamma)) + \
        g * v * v.T.dot(gamma)

    # compute step
    step = (n_samples * q / (q + 1) ** 2) * P_inv_gamma

    offset_step = step[0]
    beta_step = step[1:]

    return beta_step, offset_step


def get_step_explicit(X, y, beta, offset, lambd, q, P_inv):
    """
    Gets the MM step by explicitly calculating P^{-1} gamma
    """
    n_samples = X.shape[0]

    # compute update
    z = y * V_grad(y * (X.dot(beta) + offset), q=q) / n_samples

    gamma = X.T.dot(z) + 2 * lambd * beta
    gamma = np.insert(gamma, 0, z.sum())

    step = (n_samples * q / (q + 1) ** 2) * P_inv @ gamma

    offset_step = step[0]
    beta_step = step[1:]

    return beta_step, offset_step


def V_(u, q=1):
    """
    DWD loss function
    """
    if u <= q / (q + 1.0):
        return 1 - u
    else:
        return (1.0 / u ** q) * (q ** q / (q + 1) ** (q + 1))


# vectorized DWD loss function
V = np.vectorize(V_, excluded=['q'])


def V_grad_(u, q=1):
    """
    DWD loss function gradient
    """
    # TODO: check
    if u <= q / (q + 1.0):
        return -1
    else:
        return - (q / (u * (q + 1))) ** (q + 1)
        # return - (1.0 / u ** (q + 1)) * (q / (q + 1)) ** (q + 1)


# vectorized DWD loss gradient
V_grad = np.vectorize(V_grad_, excluded=['q'])


def dwd_obj(X, y, q, lambd, beta, offset):
    """
    DWD objective function.
    """
    return np.mean(V(y * (X.dot(beta) + offset), q=q) + lambd * beta.dot(beta))


def c_from_lambd(lambd, q, beta):
    """
    Gets the tuning paramter, C, for the SOCP formulation of DWD
    from lambda.
    """
    return ((q + 1) ** (q + 1) / q ** q) * np.linalg.norm(beta) ** (q + 1)



import cvxpy as cp
import numpy as np

from sklearn.base import BaseEstimator  # , TransformerMixin, ClassifierMixin
from sklearn.linear_model.base import LinearClassifierMixin


def solve_svm(X, y, C, sample_weight=None, solver_kws={}):
    """
    Solves soft-margin SVM problem.

    min_{beta, intercept}
    (1/n) * sum_{i=1}^n [1  - y_i * (x^T beta + intercept)] + C * ||beta||_1

    Parameters
    ----------
    X: (n_samples, n_features)

    y: (n_samples, )

    C: float
        Strictly positive tuning parameter.

    sample_weight: None, (n_samples, )
        Weights for samples.

    solver_kws: dict
        Keyword arguments to cp.solve

    Output
    ------
    beta, intercept, problem

    beta: (n_features, )
        SVM normal vector.

    intercept: float
        SVM intercept.

    problem: cp.Problem

    y_hat = np.sign(x.dot(beta) + intercept)
    """
    if sample_weight is not None:
        raise NotImplementedError

    n_samples, n_features = X.shape
    y = y.reshape(-1, 1)

    beta = cp.Variable((n_features, 1))
    intercept = cp.Variable()
    C = cp.Parameter(value=C, nonneg=True)

    # TODO: should we make this + intercept
    loss = cp.sum(cp.pos(1 - cp.multiply(y, X * beta + intercept)))
    reg = cp.norm(beta, 1)
    objective = loss / n_samples + C * reg

    problem = cp.Problem(cp.Minimize(objective))
    problem.solve(**solver_kws)

    return beta.value, intercept.value, problem


class SVM(BaseEstimator, LinearClassifierMixin):

    def __init__(self, C=1.0, solver_kws={}):
        self.C = C
        self.solver_kws = solver_kws

    def fit(self, X, y, sample_weight=None):
        self.classes_ = np.unique(y)

        self.coef_, self.intercept_, self.problem_ = \
            solve_svm(X, y, C=self.C, sample_weight=sample_weight,
                      solver_kws=self.solver_kws)

        self.coef_ = self.coef_.reshape(1, -1)
        self.intercept_ = self.intercept_.reshape(-1)

        return self



import matplotlib.pyplot as plt
import numpy as np


class ABLine2D(plt.Line2D):
    """
    Draw a line based on its slope and y-intercept. Additional arguments are
    passed to the <matplotlib.lines.Line2D> constructor.

    from https://stackoverflow.com/questions/7941226/how-to-add-line-based-on-slope-and-intercept-in-matplotlib
    """

    def __init__(self, slope, intercept, *args, **kwargs):

        # get current axes if user has not specified them
        if'axes' not in kwargs.keys():
            kwargs.update({'axes': plt.gca()})
        ax = kwargs['axes']

#         # if unspecified, get the current line color from the axes
#         if not ('color' in kwargs or 'c' in kwargs):
#             kwargs.update({'color':ax._get_lines.color_cycle.next()})

        # init the line, add it to the axes
        super(ABLine2D, self).__init__([], [], *args, **kwargs)
        self._slope = slope
        self._intercept = intercept
        ax.add_line(self)

        # cache the renderer, draw the line for the first time
        ax.figure.canvas.draw()
        self._update_lim(None)

        # connect to axis callbacks
        self.axes.callbacks.connect('xlim_changed', self._update_lim)
        self.axes.callbacks.connect('ylim_changed', self._update_lim)

    def _update_lim(self, event):
        """ called whenever axis x/y limits change """
        x = np.array(self.axes.get_xbound())
        y = (self._slope * x) + self._intercept
        self.set_data(x, y)
        self.axes.draw_artist(self)


def clf2D_slope_intercept(coef=None, intercept=None, clf=None):
    """
    Gets the slop an intercept for the separating hyperplane of a linear
    classifier fit on a two dimensional dataset.

    Parameters
    ----------
    coef:
        The classification normal vector.

    intercept:
        The classifier intercept.

    clf: subclass of sklearn.linear_model.base.LinearClassifierMixin
        A sklearn classifier with attributes coef_ and intercept_

    Output
    ------
    slope, intercept
    """

    if clf is not None:
        coef = clf.coef_.reshape(-1)
        intercept = float(clf.intercept_)
    else:
        assert coef is not None and intercept is not None

    slope = - coef[0] / coef[1]
    intercept = - intercept / coef[1]

    return slope, intercept


def get_mesh_grid(X, mesh_step=.2, excess=.5):
    """
    Returns a 2D mesh grid useful.
    """

    x_min, x_max = X[:, 0].min() - excess, X[:, 0].max() + excess
    y_min, y_max = X[:, 1].min() - excess, X[:, 1].max() + excess
    xx, yy = np.meshgrid(np.arange(x_min, x_max, mesh_step),
                         np.arange(y_min, y_max, mesh_step))
    X_mesh = np.c_[xx.ravel(), yy.ravel()]
    return X_mesh, xx, yy



from sklearn.datasets import make_blobs, make_circles
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import numpy as np

from dwd.dwd import DWD
from dwd.gen_kern_dwd import KernGDWD
from dwd.viz import ABLine2D, clf2D_slope_intercept, get_mesh_grid

cm = plt.cm.RdBu
cm_bright = ListedColormap(['#FF0000', '#0000FF'])


# sample sythetic training data
X, y = make_blobs(n_samples=200, n_features=2,
                  centers=[[0, 0],
                           [2, 2]])

# fit DWD classifier
dwd = DWD(C='auto').fit(X, y)

# compute training accuracy
tr_acc = dwd.score(X, y)

# plot points and DWD separting hyperplane
plt.scatter(X[:, 0], X[:, 1], c=y, cmap=cm_bright)
ABLine2D(*clf2D_slope_intercept(clf=dwd),
         color='black',
         label='DWD separating hyperplane')
plt.title('DWD linear classifier (tr acc = {})'.format(tr_acc))
plt.legend()
plt.savefig('dwd_sep_hyperplane.png')
plt.close()

##############
# Kernel DWD #
##############
# sample some non-linear, toy data
X, y = make_circles(n_samples=200, noise=0.2, factor=0.5, random_state=1)

# fit kernel DWD wit gaussian kernel
kdwd = KernGDWD(lambd=.1, kernel='rbf',
                kernel_kws={'gamma': 1}).fit(X, y)

# compute training accuracy
tr_acc = kdwd.score(X, y)

# visualze the decision function and training points
X_mesh, xx, yy = get_mesh_grid(X)
f = kdwd.decision_function(X_mesh)
f = f.reshape(xx.shape)
plt.contourf(xx, yy, f, cmap=cm, alpha=.5)
plt.scatter(X[:, 0], X[:, 1], c=y, cmap=cm_bright)
plt.title('DWD with Gaussian Kernel (tr acc = {})'.format(tr_acc))
plt.savefig('kern_dwd.png')
plt.close()



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.preprocessing import KernelCenterer

from dwd.kernel_utils import KernelClfMixin, KernelScaler


class KernMD(BaseEstimator, KernelClfMixin):
    def __init__(self, kernel='linear', kernel_kws={}, naive_bayes=False):
        self.kernel = kernel
        self.kernel_kws = kernel_kws

        self.naive_bayes = naive_bayes

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        self._Xfit = X  # Store K so we can compute predictions

        K = self._compute_kernel(X)

        self.dual_coef_, self.intercept_ = \
            kern_md(K, y, naive_bayes=self.naive_bayes)

        self.intercept_ = self.intercept_.reshape(-1)
        self.dual_coef_ = self.dual_coef_.reshape(1, -1)

        return self


def kern_md(K, y, naive_bayes=False):
    """
    Returns the coefficients for the kernel mean difference
    classifier.

    Parameters
    ----------
    K: array-like (n_samples, n_samples)
        Kernel matrix

    y: array-like (n_samples, )
        Vector of binary labels.

    naive_bayes: bool
        Compute naive bayes direction.

    Output
    ------
    alpha, intercept

    """
    labels = np.unique(y)
    assert K.shape[0] == K.shape[1]
    assert len(labels) == 2  # make sure binary classifier
    assert len(y) == K.shape[0]

    # center and scale K to compute Naive Bayes
    # TODO: check intercept
    if naive_bayes:
        K = KernelCenterer().fit_transform(K)
        K = KernelScaler().fit_transform(K)

    pos_ind = y == labels[1]
    neg_ind = y == labels[0]
    n_pos = sum(pos_ind)
    n_neg = sum(neg_ind)

    y_tilde = (pos_ind / n_pos) - (neg_ind / n_neg)
    alpha = K.dot(y_tilde)

    intercept = (1.0 / n_pos ** 2) * pos_ind.T.dot(K.dot(pos_ind)) - \
        (1.0 / n_neg ** 2) * neg_ind.T.dot(K.dot(neg_ind))

    return alpha, intercept



from setuptools import setup, find_packages

install_requires = ['numpy', 'cvxpy', 'sklearn', 'matplotlib']

setup(
      name='dwd',
      version='1.0.0',
      author='Iain Carmichael',
      author_email='idc9@cornell.edu',
      maintainer='Kitware Medical',
      maintainer_email='david.allemang@kitware.com',
      license='MIT',
      description='Distance weighted discrimination for Python',
      long_description=(
            'This package implements Distance Weighted Discrimination (DWD). DWD For '
            'details see (Marron et al 2007, Wang and Zou 2018). Originally implemented '
            'in Python by Iain Carmichael.'
      ),
      packages=find_packages(),
      install_requires=install_requires,
      test_suite='nose.collector',
      tests_require=['nose'],
      zip_safe=False
)



import cvxpy as cp
import numpy as np

from sklearn.base import BaseEstimator
from sklearn.linear_model.base import LinearClassifierMixin
from sklearn.utils import check_X_y
from sklearn.metrics.pairwise import euclidean_distances

from dwd.utils import pm1


class DWD(BaseEstimator, LinearClassifierMixin):

    def __init__(self, C=1.0, solver_kws={}):
        self.C = C
        self.solver_kws = solver_kws

    def fit(self, X, y, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : array-like, shape = [n_samples]
            Target vector relative to X

        sample_weight : array-like, shape = [n_samples], optional
            Array of weights that are assigned to individual
            samples. If not provided,
            then each sample is given unit weight.

        Returns
        -------
        self : object
        """
        # TODO: what to do about multi-class

        self.classes_ = np.unique(y)

        if self.C == 'auto':
            self.C = auto_dwd_C(X, y)

        # fit DWD
        self.coef_, self.intercept_, self.eta_, self.d_, self.problem_ = \
            solve_dwd_socp(X, y, C=self.C,
                           sample_weight=sample_weight,
                           solver_kws=self.solver_kws)

        self.coef_ = self.coef_.reshape(1, -1)
        self.intercept_ = self.intercept_.reshape(-1)

        return self


def solve_dwd_socp(X, y, C=1.0, sample_weight=None, solver_kws={}):
    """
    Solves distance weighted discrimination optimization problem.

    Solves problem (2.7) from https://arxiv.org/pdf/1508.05913.pdf

    Parameters
    ----------
    X: (n_samples, n_features)

    y: (n_samples, )

    C: float
        Strictly positive tuning parameter.

    sample_weight: None, (n_samples, )
        Weights for samples.

    solver_kws: dict
        Keyword arguments to cp.solve

    Output
    ------
    beta, intercept, eta, d, problem

    beta: (n_features, )
        DWD normal vector.

    intercept: float
        DWD intercept.

    eta, d:
        optimization variables.

    problem: cp.Problem

    """

    if C < 0:
        raise ValueError("Penalty term must be positive; got (C={})".format(C))

    # TODO: add sample weights
    if sample_weight is not None:
        raise NotImplementedError

    X, y = check_X_y(X, y,
                     accept_sparse='csr',
                     dtype='numeric')

    # convert y to +/- 1
    y = pm1(y)

    n_samples, n_features = X.shape

    # problem data
    X = cp.Parameter(shape=X.shape, value=X)
    y = cp.Parameter(shape=y.shape, value=y)
    C = cp.Parameter(value=C, nonneg=True)

    # optimization variables
    beta = cp.Variable(shape=n_features)
    intercept = cp.Variable()
    eta = cp.Variable(shape=n_samples, nonneg=True)

    rho = cp.Variable(shape=n_samples)
    sigma = cp.Variable(shape=n_samples)

    # objective funtion
    # TODO: check this is correct way to do sample weighting
    if sample_weight is None:
        v = np.ones(n_samples)
    else:
        v = np.array(sample_weight).reshape(-1)
        assert len(v) == n_samples
    objective = v.T @ (rho + sigma + C * eta)

    # setup constraints
    # TODO: do we need explicit SOCP constraints?
    Y_tilde = cp.diag(y)  # TODO: make sparse
    constraints = [rho - sigma == Y_tilde @ X @ beta + intercept * y + eta,
                   cp.SOC(cp.Parameter(value=1), beta)]  # ||beta||_2^2 <= 1

    # rho^2 - sigma^2 >= 1
    constraints.extend([cp.SOC(rho[i], cp.vstack([sigma[i], 1]))
                        for i in range(n_samples)])

    # solve problem
    problem = cp.Problem(cp.Minimize(objective),
                         constraints=constraints)

    problem.solve(**solver_kws)

    # d = rho - sigma
    # rho = (1/d + d), sigma = (1/d - d)/2
    d = rho.value - sigma.value

    return beta.value, intercept.value, eta.value, d, problem


def auto_dwd_C(X, y, const=100):
    """
    Automatic choice of C from Distance-Weighted Discrimination by Marron et al, 2007. Note this only is for the SOCP formulation of DWD.

    C = 100 / d ** 2

    Where d is the median distance between points in either class.

    Parameters
    ----------
    X: array-like, (n_samples, n_features)
        The input data.

    y: array-like, (n_samples, )
        The vector of binary class labels.

    const: float
        The constanted used to determine C. Originally suggested to be 100.

    """
    labels = np.unique(y)
    assert len(labels) == 2

    # pariwise distances between points in each class
    D = euclidean_distances(X[y == labels[0], :],
                            X[y == labels[1], :])

    d = np.median(D.ravel())

    return const / d ** 2


def dwd_obj(X, y, C, beta, offset, eta):
    """
    Objective function for DWD.
    """
    d = y * (X.dot(beta) + offset) + eta

    return sum(1.0 / d) + C * sum(eta)



import numpy as np
import matplotlib.pyplot as plt


def tuning_curve(X_train, y_train, X_test, y_test,
                 tuning_params, algo, solver_kws={}):
    """
    Plots tuning curve for a provided set of tuning parameters.
    """

    n_tune_values = len(tuning_params)

    train_error = np.zeros(n_tune_values)
    test_error = np.zeros(n_tune_values)

    for i, param in enumerate(tuning_params):
        beta, offset, problem = algo(X_train, y_train, param, solver_kws)

        tr_pred = lin_clf_predict(X_train, beta, offset)
        test_pred = lin_clf_predict(X_test, beta, offset)

        train_error[i] = np.mean(tr_pred != y_train)
        test_error[i] = np.mean(test_pred != y_test)

    plt.plot(tuning_params, train_error, label="Train error")
    plt.plot(tuning_params, test_error, label="Test error")
    plt.xscale('log')
    plt.legend(loc='upper left')
    plt.xlabel(r"$\lambda$", fontsize=16)
    plt.ylim(0, 1)

    return train_error, test_error


def lin_clf_predict(x, beta, offset):
    """
    Returns the linear classifier predictions for a point x.

    sign(x^T beta + offset)

    Parameters
    ----------
    x: array-like, (n_features, )
        The data point.

    beta: array-like, (n_features, )
        The classifier normal vector.

    offset: float
        The classifier offset.
    """
    return np.sign(x.dot(beta) + offset).reshape(-1)


def sample_meatballs(n_samples=1000, n_features=20, n_test=1000,
                     sep=2, p_pos=.5):
    """
    Samples binary classification problem where classes are two spherical
    Gaussians.

    Parameters
    ----------
    n_sample: int
        Sample size of training data.

    n_features: int
        Number of features.

    n_test: int
        Number of test points.

    sep: flot
        Norm of mean difference vector.

    p_pos: float
        Proportion of points in positive class

    Output
    ------
    X_train, y_train, X_test, y_test, beta_true

    """

    mean_neg = np.zeros(n_features)
    mean_pos = sep * np.ones(n_features) / np.sqrt(n_features)

    beta_true = mean_pos - mean_neg

    def sample_xy(n):
        n_pos = int(n * p_pos)
        n_neg = n - n_pos

        X_pos = np.random.multivariate_normal(mean=mean_pos, cov=np.eye(n_features), size=n_pos)
        X_neg = np.random.multivariate_normal(mean=mean_neg, cov=np.eye(n_features), size=n_neg)

        X = np.vstack([X_pos, X_neg])
        y = np.concatenate([np.ones(n_pos), -1 * np.ones(n_neg)])

        return X, y

    X_train, y_train = sample_xy(n_samples)
    X_test, y_test = sample_xy(n_test)

    return X_train, y_train, X_test, y_test, beta_true


def cv_tuning(gscv, param_name, kind='test', log=True, std=True, color='red'):
    """
    Plots the tuning curve using GridSearchCV.

    Parameters
    ----------
    gscg: from sklearn.model_selection import GridSearchCV
    """
    param_vals = [p[param_name] for p in gscv.cv_results_['params']]

    assert kind in ['train', 'test']
    st = '_{}_score'.format(kind)
    plt.plot(param_vals, gscv.cv_results_['mean' + st],
             color=color, label='cv {}'.format(kind))

    if std:
        plt.plot(param_vals,
                 gscv.cv_results_['mean' + st] + gscv.cv_results_['std' + st],
                 'b--', color=color)
        plt.plot(param_vals,
                 gscv.cv_results_['mean' + st] - gscv.cv_results_['std' + st],
                 'b--', color=color)
    if log:
        plt.xscale('log')
    plt.ylim(0, 1)
    plt.legend()
    plt.xlabel(param_name)
